# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making
# 蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
# Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.
#
# We undertake not to change the open source license (MIT license) applicable
# to the current version of the project delivered to anyone in the future.

import sys

from blue_krill.data_types.url import MutableURL


def remove_suffix(s: str, suffix: str) -> str:
    """If given string endswith `suffix`, remove the suffix and return the new string"""
    if sys.version_info >= (3, 9):
        return s.removesuffix(suffix)
    if s.endswith(suffix):
        return s[: -len(suffix)]
    return s


def remove_prefix(s: str, prefix: str) -> str:
    """If given string startswith `prefix`, remove the prefix and return the new string"""
    if sys.version_info >= (3, 9):
        return s.removeprefix(prefix)
    if s.startswith(prefix):
        return s[len(prefix) :]
    return s


def desensitize_url(url: str) -> str:
    """Obscure the password field in the url

    >>> desensitize_url("scheme://username:password@hostname:6666/path/?foo=bar")
    "scheme://username:********@hostname:6666/path/?foo=bar"
    """
    return MutableURL(url).obscure()



# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making
# 蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
# Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.
#
# We undertake not to change the open source license (MIT license) applicable
# to the current version of the project delivered to anyone in the future.

import logging
from dataclasses import asdict, dataclass
from typing import Dict, List

from blue_krill.monitoring.probe.base import Issue, VirtualProbe

try:
    import pymysql as mysql
except ImportError:
    try:
        import MySQLdb as mysql  # type: ignore  # noqa: N813
    except ImportError:
        raise ImportError("Error loading mysql module. Did you install pymysql or mysqlclient?")


logger = logging.getLogger(__name__)


@dataclass
class MySQLConfig:
    host: str
    port: int
    user: str
    password: str
    database: str


class MySQLProbe(VirtualProbe):
    """
    Usage:
        class SomeMySQLProbe(MySQLProbe):
            name: str = "some"
            config = MySQLConfig(...)
    """

    config: MySQLConfig

    def diagnose(self) -> List[Issue]:
        try:
            connection = mysql.connect(**asdict(self.config))
        except Exception:  # pylint: disable=broad-except
            logger.exception("Can't connect to MySQL server on %s", self.config.host)
            return [Issue(fatal=True, description=f"Can't connect to MySQL server on {self.config.host}")]

        try:
            with connection.cursor() as cursor:
                sql = "SELECT 1"
                cursor.execute(sql)

            connection.commit()
        except Exception as e:  # pylint: disable=broad-except
            return [Issue(fatal=True, description=f"connection to mysql<{self.name}> failed: {e}")]
        finally:
            connection.close()
        return []


def transfer_django_db_settings(django_db_settings: Dict) -> MySQLConfig:
    """transfer django db settings to MySQLConfig"""

    return MySQLConfig(
        host=django_db_settings["HOST"],
        port=int(django_db_settings["PORT"]),
        user=django_db_settings["USER"],
        password=django_db_settings["PASSWORD"],
        database=django_db_settings["NAME"],
    )



# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making
# 蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
# Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.
#
# We undertake not to change the open source license (MIT license) applicable
# to the current version of the project delivered to anyone in the future.

import logging
from typing import Optional

from cryptography.fernet import Fernet

from blue_krill.encoding import force_bytes, force_text

logger = logging.getLogger(__name__)


def get_default_secret_key():
    try:
        from django.conf import settings
    except ImportError:
        return None
    else:
        return settings.BKKRILL_ENCRYPT_SECRET_KEY


def get_default_encrypt_cipher_type():
    """获取默认的加密算法"""
    try:
        from django.conf import settings

        return getattr(settings, "ENCRYPT_CIPHER_TYPE", "FernetCipher")
    except ImportError:
        return "FernetCipher"


def encrypt_string(s: str, key: Optional[bytes] = None) -> str:
    """Shortcut function: encrypt a string into a fernet token"""
    key = key or get_default_secret_key()
    f = Fernet(key)
    value = f.encrypt(force_bytes(s))
    return force_text(value)


def decrypt_string(s: str, key: Optional[bytes] = None) -> str:
    """Shortcut function: decrypt a fernet token into a string"""
    key = key or get_default_secret_key()
    f = Fernet(key)
    value = f.decrypt(force_bytes(s))
    return force_text(value)



# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making
# 蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
# Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.
#
# We undertake not to change the open source license (MIT license) applicable
# to the current version of the project delivered to anyone in the future.

import os

MULTIPROC_ENV_NAME = "prometheus_multiproc_dir"
DEFAULT_DATA_PATH = "_prometheus_data"


def setup_multiproc_dir(data_path: str = DEFAULT_DATA_PATH):
    """Setup multi processes directory

    :param data_path: if MULTIPROC_ENV_NAME has already been configured in env vars, then the
        data_path parameter won't be used at all
    """
    dirname = os.environ.setdefault(MULTIPROC_ENV_NAME, data_path)
    try:
        os.mkdir(dirname)
    except FileExistsError:
        pass
    except IOError as e:
        print(f"unable to make prometheus data directory: {e}")



# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making
# 蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
# Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.
#
# We undertake not to change the open source license (MIT license) applicable
# to the current version of the project delivered to anyone in the future.

import logging
from binascii import Error as PaddingError

from jwt.utils import base64url_decode

logger = logging.getLogger(__name__)


def validate_jwt_token(s: str) -> bool:
    """
    Test if a given string might be a JWT token, more on JWT: https://jwt.io/introduction
    """
    parts = s.split(".")
    if len(parts) != 3:
        return False

    try:
        # JWT: {header}.{payload}.{signature}
        # The "header" and "payload" parts were always b64encoded
        base64url_decode(parts[0])
        base64url_decode(parts[1])
    except (PaddingError, ValueError):
        return False
    return True


def get_paas_service_jwt_clients():
    try:
        from django.conf import settings
    except ImportError:
        logger.exception("you should supply paas service jwt clients")
        raise
    else:
        return settings.PAAS_SERVICE_JWT_CLIENTS



# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making
# 蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
# Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.
#
# We undertake not to change the open source license (MIT license) applicable
# to the current version of the project delivered to anyone in the future.


import logging
from functools import partial
from typing import Optional

from django.db import transaction

logger = logging.getLogger(__name__)


def apply_async_on_commit(celery_task, using: Optional[str] = None, *args, **kwargs):
    """
    Apply celery task async and always ignore the task result,
    it will trigger the task on transaction commit when it is in a atomic block.
    """

    fn = partial(celery_task.apply_async, ignore_result=True, *args, **kwargs)  # noqa: B026

    connection = transaction.get_connection(using)
    if connection.in_atomic_block:
        logger.debug("trigger task %s on transaction commit", celery_task.name)
        transaction.on_commit(fn)

    else:
        logger.debug("trigger task %s immediately", celery_task.name)
        fn()


def delay_on_commit(celery_task, *args, **kwargs):
    """
    Star argument version of `apply_async_on_commit`, does not support the extra options.
    """

    apply_async_on_commit(celery_task, args=args, kwargs=kwargs)



# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making
# 蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
# Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.
#
# We undertake not to change the open source license (MIT license) applicable
# to the current version of the project delivered to anyone in the future.

from django.apps import AppConfig


class BetterLoadDataConfig(AppConfig):
    name = "blue_krill.models.better_loaddata"



# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making
# 蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
# Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.
#
# We undertake not to change the open source license (MIT license) applicable
# to the current version of the project delivered to anyone in the future.

from typing import ContextManager


class nullcontext(ContextManager):
    """Context manager that does no additional processing.
    It is a backport of feature in Python3.7 or upper.
    Used as a stand-in for a normal context manager, when a particular
    block of code is only sometimes used with a normal context manager:

    .. code-block::

        cm = optional_cm if condition else nullcontext()
        with cm:
            # Perform operation, using optional_cm if condition is True

    """

    def __init__(self, enter_result=None):
        self.enter_result = enter_result

    def __enter__(self):
        return self.enter_result

    def __exit__(self, *excinfo):
        pass



# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making
# 蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
# Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.
#
# We undertake not to change the open source license (MIT license) applicable
# to the current version of the project delivered to anyone in the future.

from typing import Optional

from django.db import models

from blue_krill.encrypt.handler import EncryptHandler


class EncryptField(models.TextField):
    """a field which will be encrypted via cryptography/fernet"""

    description = "a field which will be encrypted"

    def __init__(self, encrypt_cipher_type: Optional[str] = None, secret_key: Optional[bytes] = None, *args, **kwargs):
        super(EncryptField, self).__init__(*args, **kwargs)
        self.handler = EncryptHandler(encrypt_cipher_type=encrypt_cipher_type, secret_key=secret_key)

    def get_prep_value(self, value):
        if value is None:
            return value
        return self.handler.encrypt(value)

    def get_db_prep_value(self, value, connection, prepared=False):
        return self.get_prep_value(value)

    def from_db_value(self, value, expression, connection, context=None):
        if value is None:
            return value
        return self.handler.decrypt(value)



# -*- coding: utf-8 -*-
# TencentBlueKing is pleased to support the open source community by making
# 蓝鲸智云 - PaaS 平台 (BlueKing - PaaS System) available.
# Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
# Licensed under the MIT License (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
#     http://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
# either express or implied. See the License for the specific language governing permissions and
# limitations under the License.
#
# We undertake not to change the open source license (MIT license) applicable
# to the current version of the project delivered to anyone in the future.

import datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .ha_endpoint_pool import Endpoint


class BasicHAAlgorithm:
    """Algorithm class which cooperates with `HAEndpointPool`, it decides how
    Endpoint recover or become unhealthy.
    """

    # Max attempts before consider endpoint as "unhealthy"
    unhealthy_max_failed: int = 3
    # Unhealthy Endpoint become healthy again after `unhealthy_period`
    unhealthy_period: datetime.timedelta = datetime.timedelta(seconds=20)

    def is_unhealthy(self, endpoint: "Endpoint") -> bool:
        """Whether given endpoint is unhealthy"""
        return endpoint.failure_count >= self.unhealthy_max_failed

    def should_recover(self, endpoint: "Endpoint") -> bool:
        """Whether an unhealthy endpoint should be recovered"""
        if not endpoint.unhealthy_at:
            return False
        return (datetime.datetime.now() - endpoint.unhealthy_at) >= self.unhealthy_period
